#include "unit.h"

#include <stdlib.h> // abs()

#include <cmath>

#ifdef DEBUG
#include <iostream>
#endif // DEBUG

unit::unit(struct model * model, float x, float y) :
    energy(new gauge(model->max_energy)),
    cristal(new gauge(model->max_cristal)),
    powergem(new gauge(model->max_powergem))
{
    this->model = model;
    this->x = x;
    this->y = y;
    this->target_x = 0.0;
    this->target_y = 0.0;
    this->energy_accumulator = 0;
    this->state = automata_state::nothing;
}

unit::~unit() {
    delete(this->cristal);
    delete(this->energy);
}

void unit::draw(SDL_Renderer * renderer) {
    extern struct sprite * icon_cristal;
    extern struct sprite * icon_energy;
    extern struct sprite * icon_powergem;
    this->model->sprite.draw(renderer, this->x, this->y);
    this->powergem->draw_top(renderer, this, icon_powergem, 0);
    this->cristal->draw_top(renderer, this, icon_cristal,this->powergem->value);
    this->energy->draw_bot(renderer, this, icon_energy, 0);
}

bool unit::can_build(struct model * model) {
    return(this->energy->value  >= model->energy_cost
        && this->cristal->value >= model->cristal_cost);
}

float unit::distance() {
    return(this->distance(this->target_x, this->target_y));
}

float unit::distance(struct unit * unit) {
    return(this->distance(unit->x, unit->y));
}

float unit::distance(float x, float y) {
    float X = abs(this->x - x);
    float Y = abs(this->y - y);
    float dist = sqrt(X*X+Y*Y);
    // float dist = X+Y; // Manhattan distance
    return(dist);
}

void unit::set_position(float x, float y) {
    if(x >= 0 && x < SCREEN_WIDTH && y >= 0 && y < SCREEN_HEIGHT) {
        this->x = x;
        this->y = y;
    }
}

void unit::move(float dx, float dy,
            struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->model->speed > 0) {
        float nx = this->x + dx * this->model->speed;
        float ny = this->y + dy * this->model->speed;
        int mx = nx/SQUARE_SIZE;
        int my = ny/SQUARE_SIZE;
        if(grid[mx][my].is_passable()) {
            this->set_position(nx, ny);
        } else if(grid[mx][(int)this->y/SQUARE_SIZE].is_passable()) {
            this->set_position(nx, this->y);
        } else if(grid[(int)this->x/SQUARE_SIZE][my].is_passable()) {
            this->set_position(this->x, ny);
        }
    }
}

void unit::move_toward(float x, float y,
            struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->distance(x, y) > SQUARE_SIZE/3) {
        float X = x - this->x;
        float Y = y - this->y;
        float dist = this->distance(x, y);
        float dx = X/dist;
        float dy = Y/dist;
#ifdef DEBUG
    if(this->model->name == "midro") {
        std::cout << this << " move toward " << x << "~" << y << " from " << this->x << "~" << this->y << " at distance " << dist << std::endl;
    }
#endif // DEBUG
        this->move(dx, dy, grid);
    }
}

void unit::mine(struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->model->drill_power > 0 && !this->energy->is_empty()) {
        this->energy->decrease(1);
        for(int dx = -1; dx <= +1; dx++) {
            for(int dy = -1; dy <= +1; dy++) {
                int nx = (this->x + dx*SQUARE_SIZE/2) / SQUARE_SIZE;
                int ny = (this->y + dy*SQUARE_SIZE/2) / SQUARE_SIZE;
                if(nx >= 0 && nx < GRID_WIDTH && ny >= 0 && ny < GRID_HEIGHT) {
                    grid[nx][ny].mine(this);
                }
            }
        }
    }
}

void unit::build(struct model * model, float x, float y,
                 std::forward_list<struct unit *> * units) {
    if(this->can_build(model)) {
        this->energy->decrease(model->energy_cost);
        this->cristal->decrease(model->cristal_cost);
        units->push_front(new unit(model, x, y));
    }
}

void unit::cycle(std::forward_list<struct unit *> * units,
                 struct square grid[GRID_WIDTH][GRID_HEIGHT]) {

    if(this->model->generate_energy > 0 && !this->energy->is_full())
        this->generate_energy();

    if(this->model->collect_energy)
        this->collect_energy(units);

    if(this->model->collect_cristal)
        this->collect_cristal(units);

    if(this->model->collect_powergem)
        this->collect_powergem(units);

    if(this->model->autobuild != NULL)
        this->build(this->model->autobuild, this->x, this->y, units);

    if(this->model->name == "midro") {
        switch(this->state) {
        case automata_state::nothing :
            if(this->energy->is_empty()) {
                this->state = automata_state::refuel;
                this->target_nearest_energen(units);
#ifdef DEBUG
                std::cout << this << " : nothing -> refuel " << this->target_x << "~" << this->target_y << std::endl;
#endif
            } else {
                this->state = automata_state::mine;
                this->target_nearest_cristal(grid);
#ifdef DEBUG
                std::cout << this << " : nothing -> mine" << this->target_x << "~" << this->target_y<< std::endl;
#endif
            }
            break;
        case automata_state::refuel :
            if(this->energy->is_full()) {
                this->state = automata_state::mine;
                this->target_nearest_cristal(grid);
#ifdef DEBUG
                std::cout << this << " : refuel -> mine" << this->target_x << "~" << this->target_y<< std::endl;
#endif
            }
            break;
        case automata_state::mine :
            if(this->count_surrounding_rock(grid) >= 2
                    || this->distance() < SQUARE_SIZE/2) {
                this->mine(grid);
            }
            if(this->cristal->is_full()) {
                this->state = automata_state::drop;
                this->target_nearest_builder(units);
#ifdef DEBUG
                std::cout << this << " : mine -> drop" << this->target_x << "~" << this->target_y<< std::endl;
#endif
            } else if(this->energy->is_empty()) {
                this->state = automata_state::refuel;
                this->target_nearest_energen(units);
#ifdef DEBUG
                std::cout << this << " : mine -> refuel" << this->target_x << "~" << this->target_y<< std::endl;
#endif
            }
            break;
        case automata_state::drop :
            if(this->cristal->is_empty()) {
                this->state = automata_state::nothing;
#ifdef DEBUG
                std::cout << this << " : drop -> nothing" << this->target_x << "~" << this->target_y<< std::endl;
#endif
            }
            break;
        }
    }

    if(this->state != automata_state::nothing) {
        this->move_toward(this->target_x, this->target_y, grid);
    }
}

void unit::generate_energy() {
    if(!this->powergem->is_empty()) {

        this->energy_accumulator +=
                this->model->generate_energy * this->powergem->value;

        if(this->energy_accumulator >= ENERGY_GENERATOR_THRESHOLD) {
            this->energy->increase(1);
            this->energy_accumulator -= ENERGY_GENERATOR_THRESHOLD;
        }
    }
}

void unit::collect_energy(std::forward_list<struct unit *> * units) {
    for(struct unit * unit : *units) {
        if(
            unit->model->give_energy &&
            this->model != unit->model &&
            this->distance(unit) < COLLECT_DISTANCE
        ) {
            this->energy->gather(unit->energy);
        }
    }
}

void unit::collect_cristal(std::forward_list<struct unit *> * units) {
    for(struct unit * unit : *units) {
        if(
            unit->model->give_cristal &&
            this->model != unit->model &&
            this->distance(unit) < COLLECT_DISTANCE
        ) {
            this->cristal->gather(unit->cristal);
        }
    }
}

void unit::collect_powergem(std::forward_list<struct unit *> * units) {
    for(struct unit * unit : *units) {
        if(
            unit->model->give_powergem &&
            this->model != unit->model &&
            this->distance(unit) < COLLECT_DISTANCE
        ) {
            this->powergem->gather(unit->powergem);
        }
    }
}

struct unit * unit::nearest_unit(struct model * model,
            std::forward_list<struct unit *> * units) {
    struct unit * nearest = NULL;
    float minimum = SCREEN_WIDTH * 100;
    for(struct unit * it : *units) {
        if(it->model == model) {
            float dist = this->distance(it);
            if(dist < minimum) {
                minimum = dist;
                nearest = it;
            }
        }
    }
    return(nearest);
}

struct unit * unit::nearest_unit(std::string model_name,
            std::forward_list<struct unit *> * units) {
    struct unit * nearest = NULL;
    float minimum = SCREEN_WIDTH * 100;
    for(struct unit * it : *units) {
        if(it->model->name == model_name) {
            float dist = this->distance(it);
            if(dist < minimum) {
                minimum = dist;
                nearest = it;
            }
        }
    }
    return(nearest);
}

void unit::target_nearest_cristal(struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    bool found = false;
    unsigned int nearest_x = 0;
    unsigned int nearest_y = 0;
    float minimum = SCREEN_WIDTH * 100;
    for(unsigned int x=0; x<GRID_WIDTH; x++) {
        for(unsigned int y=0; y<GRID_HEIGHT; y++) {
            if(grid[x][y].cristal > 0) {
                float nx = x*SQUARE_SIZE + SQUARE_SIZE/2;
                float ny = y*SQUARE_SIZE + SQUARE_SIZE/2;
                float dist = this->distance(nx, ny);
                if(dist < minimum) {
                    found = true;
                    minimum = dist;
                    nearest_x = nx;
                    nearest_y = ny;
                }
            }
        }
    }
    if(found) {
        this->target_x = nearest_x;
        this->target_y = nearest_y;
    }
}

void unit::target_nearest_energen(std::forward_list<struct unit *> * units) {
    struct unit * target = this->nearest_unit("enerstash", units);
    if(target == NULL || target->energy->is_empty()) {
        target = this->nearest_unit("energen", units);
        if(target == NULL || target->powergem->is_empty()) {
            target = this->nearest_unit("midripod", units);
        }
    }
    if(target != NULL) {
        this->target_x = target->x;
        this->target_y = target->y;
    }
}

void unit::target_nearest_builder(std::forward_list<struct unit *> * units) {
    struct unit * target = this->nearest_unit("factory", units);
    if(target == NULL || target->cristal->is_full()) {
        target = this->nearest_unit("midripod", units);
    }
    if(target != NULL) {
        this->target_x = target->x;
        this->target_y = target->y;
    }
}

unsigned int unit::count_surrounding_rock(struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    unsigned int count = 0;
    for(int dx = -1; dx <= +1; dx++) {
        for(int dy = -1; dy <= +1; dy++) {
            int nx = (this->x + dx*SQUARE_SIZE/2) / SQUARE_SIZE;
            int ny = (this->y + dy*SQUARE_SIZE/2) / SQUARE_SIZE;
            if(nx >= 0 && nx < GRID_WIDTH && ny >= 0 && ny < GRID_HEIGHT) {
                if(grid[nx][ny].rock > 0) {
                    count++;
                }
            }
        }
    }
    return(count);
}
